"""
Module providing perturbation functions under a common interface
"""
from art.attacks.poisoning.perturbations.image_perturbations import (
    add_pattern_bd,
    add_single_bd,
    insert_image,
)
