"""
This module implements the fast generalized subset scan based detector.
"""
from art.defences.detector.evasion.subsetscanning.scanningops import ScanningOps
from art.defences.detector.evasion.subsetscanning.scanner import Scanner
from art.defences.detector.evasion.subsetscanning.detector import SubsetScanningDetector
