"""
Module providing inference attacks.
"""
from art.attacks.inference import attribute_inference
from art.attacks.inference import membership_inference
from art.attacks.inference import model_inversion
