"""
Module implementing detector-based defences against evasion attacks.
"""
from art.defences.detector.evasion import subsetscanning

from art.defences.detector.evasion.detector import BinaryInputDetector, BinaryActivationDetector
