"""
Module providing extraction attacks under a common interface.
"""
from art.attacks.extraction.functionally_equivalent_extraction import FunctionallyEquivalentExtraction
from art.attacks.extraction.copycat_cnn import CopycatCNN
from art.attacks.extraction.knockoff_nets import KnockoffNets
